#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#define likely(x)   __builtin_expect((x),1)
#define unlikely(x) __builtin_expect((x),0)

#define TRACE 1

/* define 64bit registers; change arrays for larger one */
typedef union reg_s {
	unsigned long long u64;
	signed long long s64;
	unsigned u32[2];
	int s32[2];
	unsigned short u16[4];
	signed short s16[4];
	unsigned char u8[8];
	signed char s8[8];
} reg_t;

reg_t reg[64];
unsigned stamp[64];
unsigned now; 		/* actual timestamp */

/* write port reservation circular FIFO; size must be greater than
   highest instruction latency; it must also be power of 2 */
#define WPR_SZ 64
#define WPORTS 2	/* how many wports we have */

/* each item is no of reserved write port at that time */
unsigned char wpr[WPR_SZ];
int wpr_now = 0;	/* index of current slot */

/* move to time in future */
void __inline__ shift_time(unsigned moveto)
{
	int i,diff = moveto - now;
	assert(diff <= WPR_SZ);
	
	/* clear diff slots */
	i = wpr_now + diff;
	if (i <= WPR_SZ) memset(wpr+wpr_now,0,diff);
	else {
		i &= WPR_SZ-1;
		memset(wpr+wpr_now,0,WPR_SZ-wpr_now);
		memset(wpr,0,i);
	}
	wpr_now = i;

	/* change actual time and check for overflow */
	if ((now = moveto) > 0x10000000) {
		for (i=0;i<64;i++) stamp[i] = stamp[i] < moveto ? 0 : stamp[i] - moveto;
		now = 0;
	}
}

/* looks from given time onward for free writeport(s); returns stall cycles */
unsigned __inline__ alloc_wport(unsigned wr,unsigned latency)
{
	int i;
	int howmany = wr ? (wr > 63 ? 2 : 1) : 0; /* TODO: be more generic */
	if (!wr) return 0; 			/* no write needed */
	if (!(wr & 0x3f)) howmany--;	/* r0 write */
	
	for (i = latency; i < WPR_SZ; i++) {
		int p = (wpr_now + i) & (WPR_SZ-1);
		if (WPORTS - wpr[p] >= howmany) {
			wpr[p] += howmany;	/* schedule our write */
			return i-latency;
		}
	}
	abort();	/* Fatal: we can't place write ! Probably WPR_SZ too low. */
}

#define OBIT(N) ((opc>>N) & 1)
#define OBIT2(N) ((opc>>N) & 3)
#define OSZ_CNT 4
#define OSZ(N) ((opc>>N) & 3) /* size in log(bytes) - 0=1B,1=2B,2=4B,3=8B...OSZ_CNT-1 */
#define REG1 (opc & 0x3f)
#define REG2 ((opc >> 6) & 0x3f)
#define REG3 ((opc >> 12) & 0x3f)
#define OCASE1(I) case OPCODE_##I

/* include opnumber defs here */
#include "opnums.h"

/* simulates opcode; returns number of stall cycles */
int simulate_insn(unsigned opc)
{
	int latency = 1,delay,wdelay;
	unsigned mstamp,u,rd = (opc >> 6) & 0xfff;
   	unsigned wr = opc & 0x3f;	/* default is 2R1W */
	
	switch(opc>>24) {
		/* ------------------ ISA semantic is defined here --------------- */
#define SMMX(C,U) asm (C:"=y" (reg[REG1].u64):"0" (reg[REG2].u64),"y" (reg[REG3].u64) U)
#define SMMXR(C,U) asm (C:"=y" (reg[REG1].u64):"0" (reg[REG3].u64),"y" (reg[REG2].u64) U)
#define SMMX2(C,U...) asm (C:"=y" (reg[REG1^1].u64):"y" (reg[REG2].u64),"y" (reg[REG3].u64),"0" (reg[REG1^1].u64) U)
#define SMMX3(C,U...) asm (C:"=y" (reg[REG1^1].u64):"y" (reg[REG2].u64),"y" (reg[REG3].u64),"0" (reg[REG1^1].u64),"y" (reg[REG1].u64) U)
#define SMMX4(C,U...) asm (C:"=y" (reg[REG1].u64):"y" (reg[REG2].u64),"y" (reg[REG3].u64),"0" (reg[REG1^1].u64),"y" (reg[REG1].u64) U)
		/* ROP2; MMX is mandatory to get reasonable timing */
		OCASE1(AND): SMMX("pand %2,%0",); break;
		OCASE1(ANDN):SMMXR("pandn %2,%0",); break;
		OCASE1(XOR): SMMX("pxor %2,%0",); break;
		OCASE1(OR):  SMMX("por %2,%0",); break;
		/* How to get (s64)-1 ? pcmpeqb r,r gives us 64 ones.
		   we need it to implement negation of result by xor. */
		OCASE1(NOR): SMMX("pcmpeqb %%mm2,%%mm2\npxor %2,%0\npxor %%mm2,%0",:"mm2"); break;
		OCASE1(XNOR):SMMX("pcmpeqb %%mm2,%%mm2\npor %2,%0\npxor %%mm2,%0",:"mm2"); break;
		OCASE1(ORN): SMMX("pcmpeqb %%mm2,%%mm2\npandn %2,%0\npxor %%mm2,%0",:"mm2"); break;
		OCASE1(NAND):SMMX("pcmpeqb %%mm2,%%mm2\npand %2,%0\npxor %%mm2,%0",:"mm2"); break;
		OCASE1(MUX): SMMX3("movq %2,%%mm2\n pand %1,%%mm2\nmovq %1,%0\n pandn %4,%0\n"
			   			"por %%mm2,%0",:"mm2");
					 rd |= wr << 12; wr ^= 1; /* 3R1W */
					 break;

		/* ASU; MMX is used for smaller chunk ops; FIXME:assume signed saturation */
		OCASE1(ADD): 
			latency = 2;
			switch (OSZ(22)) {
				case 0: SMMX("paddb %2,%0",); 
						if (unlikely(OBIT(18))) { /* carry */
							reg[REG1^1].u64 = 0x0101010101010101; /* to mask and negate result */
							SMMX3("movq %1,%%mm2\npaddusb %2,%%mm2\nmovq %3,%%mm4\nmovq %4,%0\n"
									"pcmpeqb %%mm2,%0\npandn %%mm4,%0\n",:"mm2","mm4");
						}
						if (unlikely(OBIT(19))) { /* saturate */
							SMMX("paddusb %2,%0",);
						}
						latency = 1;
						break;
				case 1: SMMX("paddw %2,%0",); 
						if (unlikely(OBIT(18))) { /* carry */
							SMMX3("movq %4,%%mm2\n movq %1,%%mm3\n pxor %2,%%mm3\n"
									"pandn %%mm3,%%mm2\n movq %1,%0\n pand %2,%0\n por %%mm2,%0\n"
									"psrlw $15,%0",:"mm2","mm3"); 
						}
						if (unlikely(OBIT(19))) { /* saturate */
							SMMX("paddusw %2,%0",);
						}
						break;
				case 2: SMMX("paddd %2,%0",); 
						if (unlikely(OBIT(18))) { /* carry */
							SMMX3("movq %4,%%mm2\n movq %1,%%mm3\n pxor %2,%%mm3\n"
									"pandn %%mm3,%%mm2\n movq %1,%0\n pand %2,%0\n por %%mm2,%0\n"
									"psrld $31,%0",:"mm2","mm3"); 
						}
						if (unlikely(OBIT(19))) { /* saturate */
							SMMX4("movq %4,%%mm2\n movq %1,%%mm3\n pxor %2,%%mm3\n"
									"pandn %%mm3,%%mm2\n movq %1,%0\n pand %2,%0\n por %%mm2,%0\n"
									"psrad $31,%0\n por %4,%0",:"mm2","mm3"); 
						}
						break;
				case 3: reg[REG1].u64 = reg[REG2].u64 + reg[REG3].u64;
						if (unlikely(OBIT2(18))) { /* detect overflow if needed */
							unsigned int ovr = ((reg[REG2].u32[1]&reg[REG3].u32[1])|
								((reg[REG2].u32[1]^reg[REG3].u32[1])&~reg[REG1].u32[1]))>>31;
							if (OBIT(18)) reg[REG1^1].u64 = ovr; /* store carry */
							if (OBIT(19) && ovr) /* saturate */
								reg[REG1].u64 = 0xffffffffffffffff;
						}
						break;
				default: abort();
			}
		    if (unlikely(OBIT(18))) wr |= (wr^1) << 6; /* 1W -> 2W */
		    break;

		/* Moves */
		OCASE1(MOVE): 
			latency = 0; // TODO: partial moves and other conditions
			if (!reg[REG3].u64) reg[REG1].u64 = reg[REG2].u64;
			break;

		
		/* ------------------ End of ISA definitions --------------------- */
		default:
			// TODO: exception
			fprintf(stderr,"Unhandled opcode 0x%X\n",opc);
			exit(1);
	}
	/* side efects are done, let's detect RAW collision */
	/* NOTE: at this point we are at now+1 */
	mstamp = now+1;
	for (;rd;rd >>= 6) if (stamp[rd & 0x3f] > mstamp) mstamp = stamp[rd & 0x3f];
	for (u=wr;u;u >>= 6) if (stamp[u & 0x3f] > mstamp) mstamp = stamp[u & 0x3f];

	delay = mstamp - now - 1; 	/* how long we have to wait for prev. result */
	if (delay)
		printf("RAW/WAW stall %d cycles !\n",delay);
	
	shift_time(mstamp);			/* move to time when insn can be scheduled */

	/* try to allocate write port(s) in actual time */
	wdelay = alloc_wport (wr,latency);
	if (wdelay) {
		printf("Write port stall %d cycles !\n",wdelay);
		shift_time(now+wdelay);
	}

	/* mark written registers as busy */
	for (u=wr;u;u >>= 6) stamp[u & 0x3f] = now + latency + 1;

	/* handle r0 */
	reg[0].u64 = 0; stamp[0] = 0;
	
	return delay;
}

void printfifo()
{
	int i;
	printf("FIFO wports assignments:\n");
	for (i=0;i<WPR_SZ-10;i+=10) printf("         |");
	printf("\n");
	for (i=0;i<WPR_SZ;i++) {
		int p = (wpr_now + i) & (WPR_SZ-1);
		printf("%1d",wpr[p]);
	}
	printf("\n\n");
}
	
void printregs()
{
	int i;
	printf("Now: %4u\n",now);
	for (i=0;i<8;i++) {
		printf("r%-2d [%4u]: 0x%016LX\n",i,stamp[i],reg[i].u64);
	}
}

#define OP(CODE,SZ,FL,R3,R2,R1) (((OPCODE_##CODE)<<24)|((SZ)<<22)|((FL)<<18)|((R3)<<12)|((R2)<<6)|R1)
int main()
{
	int i;
	//                              S0rrsstt
	unsigned prg[] = {
		OP(ADD,3,0,1,2,5),
		OP(AND,3,0,1,2,6),
		OP(MOVE,3,0,0,3,7),
		OP(AND,3,0,1,2,6),
		OP(XOR,3,0,1,6,6),
		0
	},*p=prg;
	


	reg[1].u64 = 0xff100005ffffffff;
	reg[2].u64 = 0x0301000500000001;
	reg[3].u64 = 0xffffff00f000fff0;
	printregs(); printfifo();
	while(*p) {
		simulate_insn(*(p++));
		printregs(); printfifo();
	}
	return 0;
}
